/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 14.6 (Perfect Random Walk Metropolis Sampler, p. 382)
 *
 * this file: algo14_06.c
 *
 * For a demo how to use this algorithm see the file "demo_algo14_06.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"




#define S (rp->s)
#define U1 (rp->u1)
#define U2 (rp->u2)

struct randpairs{
  int      n;/*maximal number of random pairs*/
  double *u1;/*pointer to arrays of U(0,1) random  variables*/
  double *u2;/*pointer to array of U(-s,s) random  variables*/
  double s;/*parameter for u2*/
};


void ini_randpairs(int n, double s, struct randpairs *rp)
     /*initializes the structure randpairs*/
{
 rp->n=n;
 rp->s=s;
 rp->u1=malloc(n*sizeof(double));
 rp->u2=malloc(n*sizeof(double));
}

void free_randpairs(struct randpairs *rp)
{ free(rp->u1);
  free(rp->u2);
}

int add_randpairs(int t1,int t2, struct randpairs *rp)
/*generates new random numbers for randpair
  between t1 and t2*/
{ int i;
 if(t2>rp->n)
   { printf("array for randpairs too small, n=%d is not enough\n",rp->n);
   return(0);
     }
 for(i=t1;i<t2;i++) 
 { rp->u1[i]=uniform();
   rp->u2[i]=(2.*S)*(uniform()-0.5);
 }
 return 1;
}

void onestep(double (*pdf)(double),int t, double x[2], struct randpairs *rp, double bl, double br)
/* simulates one step of the markov chain
   t   ... time after that step
   x[0]... Minimal state after step
   x[1]... maximal state after step
*/

{ int i;
  double xn[2];/*propsed new values*/

          for(i=0;i<2;i++){ 
/*7+10*/    xn[i]=((int)((x[i]+S-U2[t])/(2.*S)))*2.*S+U2[t];
/*8+11*/  if(U1[t]*(*pdf)(x[i]) <= (*pdf)(xn[i])) 
/*9+12*/    x[i]=xn[i];
  }
}

/*************************************/

double generator14_06( double (*pdf)(double), double bl, double br, double s)
/*
 * Algorithm 14.6 (Perfect Random Walk Metropolis Sampler, p. 382)
 *
 * pdf ... function pointer to the density, which must be monotone
 * bl  ... left border of the domain
 * br  ... right border of the domain
 * s  ... random walk proposal is U(-s,s)
 *
 */
{ struct randpairs rp;
  double x[2];
  int n=30000, t0,t0s,t0a,t,startt0=4;

       x[0] = bl;
       x[1] = br;
/*1*/  t0s=startt0;
       t0a=0;  
/*2*/  ini_randpairs(n,s,&rp);  
/*3*/  for(t0=t0s;/*14*/(fabs(x[0]-x[1])>1.e-15);){
         if(!add_randpairs(t0a,t0,&rp)) return(0.5*(x[0]+x[1]));
/*4*/    x[0]=bl;
         x[1]=br;
/*5*/    for(t=t0-1;t>=0;t--) 
           onestep(pdf,t,x,&rp,bl,br);
         t0a=t0;
/*13*/	 t0*=2;
       }
       free_randpairs(&rp);
       /* printf("end\n"); */
       if(fabs(x[0]-x[1])<1.e-15)
/*15*/	 return x[0];
       return -1.;
}


